#include "getopti.h"
#include <string.h>
#include <stdlib.h>
#include "headio.h"
#include "Headped.H"
#include "util.h"

#define MAXNGEN 256

static char version[]="0.34 (20030127) JC Loredo-Osti and K Morgan.";

extern "C"{
    void initpedc(int,int,int,int,int);
    void outputpedc(FILE *, char *, int);
    void mngrcoords(void);
    void annealcoords(int, float, float);
    int  readlist(FILE *, int,int,int,char *,char *, char *);
}

void usage(char *name){
  fprintf(stderr,"\n     Usage: %s [options] <file>\n\n"\
	  "   Transforms a list in a pedigree coordinates file.\n"\
	  "   List columns are:\n"\
	  "     <id> <father> <mother> [sex] [affection_status] [label]\n"\
	  "         <> and [] mean mandatory and optional fields, respectively.\n"\
	  "Options:\n",name);
  fprintf(stderr,"\t{ -o | --output }<file>     Output file\n"\
	  "\t{ -s | --sex }              Input file includes sex\n"\
	  "\t{ -a | --affection }        Input file includes affection status\n"\
	  "\t{ -l | --labels }           Last column of input file holds labels\n"\
	  "\t{ -m | --minimize }         Minimize the square of line distances\n"\
	  "\t                                 by simulated annealing\n");
  fprintf(stderr,"\t{ -I | --iterations }<n>    Iterations for annealing\n"\
	  "\t{ -T | --temperature }<f>   Annealing starting temperature\n"\
	  "\t{ -C | --cooling }<f>       Annealing cooling factor\n"\
	  "\t                            RGB color for the affected status:\n"\
	  "\t{ -N | --normal }<r,g,b>         0 < r,g,b <= 255\n"\
	  "\t{ -A | --affected }<r,g,b>       without spaces between numbers\n"\
	  "\t{ -U | --unknown }<r,g,b>\n"\
	  "\t{ -d | --default }          Show default values\n"\
	  "\t{ -v | --version }\n");
}

void defaults(char *name){
  fprintf(stderr,"\n  %s options have the following default values or actions:\n\n"\
	  "\t        Option \t Value\n\n"\
	  "\t        output \t stdout\n"\
	  "\t           sex \t no sex information\n"\
	  "\t     affection \t no affection status information\n"\
	  "\t        labels \t id number taken as label\n"\
	  "\t      minimize \t no minimization\n"\
	  "\t    iterations \t 1000\n",name);
  fprintf(stderr,"\t   temperature \t 15.0\n"\
	  "\t       cooling \t 0.9\n"\
	  "\t               \t RGB color for the affected status:\n"\
	  "\t        normal \t 255,255,255   (white)\n"\
	  "\t      affected \t 25,25,112     (midnight blue)\n"\
	  "\t       unknown \t 255,222,173   (Navajo white)\n"\
	  "\n");
  exit(0);
}

int main(int argc, char **argv)
{
  int i,  count, annealing=0;
  int sexes = 0, label = 0;
  char filename[maxline], pedname[maxstr];
  FILE *infile, *outfile;
  int kk, tmp=0;
  int nit=1000;
  float boil;
  float cool;
  int affection = 0;
  char normal[65];
  char affected[65];
  char unknown[65];
  
  static struct option _options_[] = 
  {    
    {"output", 1, 'o'},
    {"sex", 0, 's'},
    {"affection", 0, 'a'},
    {"label", 0, 'l'},
    {"minimize",0,'m'},
    {"iterations",1,'I'},
    {"temperature",1,'T'},
    {"cooling",1,'C'},
    {"normal",1,'N'},
    {"affected",1,'A'},
    {"unknown",1,'U'},
    {"default",1,'d'},
    {"version",0, 'v'},
    {0, 0, 0}
  };
  boil = 15.0;
  cool = 0.9;
  pedname[0]='\0';
  strcpy(normal,"255,255,255");
  strcpy(affected,"25,25,112");
  strcpy(unknown,"255,222,173");
  GetOpt opto(argc, argv, "salmo:I:T:C:vN:A:U:d", _options_);
  while (1){

    kk = opto();
    if (kk == -1)
      break;
      
    switch (kk)
      {
      case 'I':
	if(opto.optarg){
	  tmp = atoi(opto.optarg);
	  if(tmp > 0) nit = tmp;
	}
	break;
      case 'T':
	if(opto.optarg){
	  boil = atof(opto.optarg);
	}
	break;
      case 'C':
	if(opto.optarg){
	  cool = atof(opto.optarg);
	}
	break;
      case 'N':
	if(opto.optarg){
	  strcpy(normal,opto.optarg);
	}
	break;
      case 'A':
	if(opto.optarg){
	  strcpy(affected,opto.optarg);
	}
	break;
      case 'U':
	if(opto.optarg){
	  strcpy(unknown,opto.optarg);
	}
	break;
      case 'm':
	annealing = 1;
	break;
      case 'd':
	defaults(argv[0]);
	break;
      case 'a':
	affection = 1;
	break;
      case 'l':
	label = 1;
	break;
      case 's':
	sexes = 1;
	break;
      case 'o':
	if(! opto.optarg){ usage(argv[0]); exit(1);}
	strcpy(pedname,opto.optarg);
	break;
      case 'v':
	fprintf(stdout, 
		"Version %s.\n", version);
	exit(0);
      default :
	usage(argv[0]);
	exit(1);
      }
  }

  if(opto.optind < argc)
    strcpy(filename, argv[opto.optind++]);
  else{
    usage(argv[0]);
    exit(1);
  }

  initpedc(15000,15000,15000,15000,30000);
  if ( (infile = fopen(filename,"r")) == NULL ){
	fprintf(stderr,"\n Failure to open intput file %s.\n",filename); 
	exit(1);
  }
  else{
    count = readlist(infile,sexes,affection,label,normal,affected,unknown);
    fclose(infile);
    for (i=1; i<=indpt; i++) indstack[i].info.flag = UP;
    for (i=1; i<=marpt; i++) marstack[i].info.flag = UP;
    mngrcoords();
    if (annealing)   annealcoords(nit,boil,cool);
    if(pedname[0] != '\0'){
      if ( (outfile = fopen(pedname,"w")) == NULL ){ 
	fprintf(stderr,"\n Failure to open output file %s.",pedname); 
	fprintf(stderr,"\n Check write status of directory.");
	exit(1);
      }
      else{
	outputpedc(outfile,pedname,0);
	fclose(outfile);
      }    
    }
    else{
      outputpedc(stdout,filename,0);
      fflush(stdout);
    }
  }
  if(pedname[0] == '\0')  fprintf(stderr,"\n");
  exit(0);
}
